@@ -1,7 +1,6 @@
 import os
 import sys
 import logging
-import hashlib
 import grpc
 import json
 import threading
@@ -11,7 +10,6 @@ import shlex
 
 from typing import Tuple, Union
 
-from tempfile import gettempdir
 from crontab import CronTab
 from dotenv import dotenv_values
 from io import StringIO
@@ -36,11 +34,7 @@ SPECIALS = {"reboot":   '@reboot',
 
 METADATA = (('containerd-namespace', 'k8s.io'),)
 
-
-def hashsum(str):
-    md5 = hashlib.md5()
-    md5.update(str.encode('utf-8'))
-    return md5.hexdigest()
+FIFO_DIR = '/tmp/containerd-fifo'
 
 
 class ReadTask(threading.Thread):
@@ -74,12 +68,17 @@ class ReadTask(threading.Thread):
         return self.output
 
 
-def mkfifo(exec_id: str):
-    fifo = gettempdir() + '/' + exec_id
+def rmfifo(exec_id: str):
+    fifo = FIFO_DIR + '/' + exec_id
     try:
         os.unlink(fifo)
     except:
         pass
+    return fifo
+
+
+def mkfifo(exec_id: str):
+    fifo = rmfifo(exec_id)
     os.mkfifo(fifo)
     return fifo
 
@@ -92,7 +91,7 @@ def get_container_spec(channel: grpc.Channel, container_id: str):
 
 
 def run_command(channel: grpc.Channel, container_id: str, args, timeout=5) -> Tuple[int, str]:
-    exec_id = "exec-" + hashsum(container_id + ': ' + str(args))
+    exec_id = "exec-" + os.urandom(16).hex()
     stdout = mkfifo(exec_id)
     read_task = ReadTask(stdout)
 
@@ -139,7 +138,9 @@ def run_command(channel: grpc.Channel, container_id: str, args, timeout=5) -> Tu
     except:
         exit_status = 1
 
-    return exit_status, read_task.result()
+    result = read_task.result()
+    rmfifo(exec_id)
+    return exit_status, result
 
 
 def run_schedule(channel: grpc.Channel, container_id: str, args):
@@ -149,50 +150,55 @@ def run_schedule(channel: grpc.Channel, container_id: str, args):
     return exit_code
 
 
-def get_os_id(channel: grpc.Channel, container_id: str):
-    exit_code, output = run_command(
-        channel, container_id, ["cat", "/etc/os-release"])
-    if exit_code != 0:
-        return
-    release = dotenv_values(stream=StringIO(output.decode('utf8')))
-    return release['ID']
+# def get_os_id(channel: grpc.Channel, container_id: str):
+#     exit_code, output = run_command(
+#         channel, container_id, ["cat", "/etc/os-release"])
+#     if exit_code != 0:
+#         return
+#     release = dotenv_values(stream=StringIO(output.decode('utf8')))
+#     return release['ID']
 
 
-def get_alpine_crontab(channel: grpc.Channel, container_id: str) -> Tuple[str, Union[str, bool]]:
+def get_current_user(channel: grpc.Channel, container_id: str) -> str:
     exit_code, output = run_command(
         channel, container_id, ['whoami'])
-    if exit_code != 0:
-        return '', None
-    user = output.decode('utf8').replace('\n', '')
+    if exit_code == 0:
+        return output.decode('utf8').replace('\n', '')
+    return 'root'
+
+
+def get_user_crontab(channel: grpc.Channel, container_id: str, user: str) -> CronTab:
     exit_code, output = run_command(
         channel, container_id, ['cat', '/etc/crontabs/{user}'.format(user=user)])
     if exit_code != 0:
-        return '', None
-    return output.decode('utf8').replace('\t', ' '), user
+        return None
+    tab = output.decode('utf8').replace('\t', ' ')
+    return CronTab(tab=tab, user=user)
 
 
-def get_debian_crontab(channel: grpc.Channel, container_id: str) -> Tuple[str, Union[str, bool]]:
+def get_system_crontab(channel: grpc.Channel, container_id: str) -> CronTab:
     exit_code, output = run_command(channel, container_id, ["/bin/sh", "-c",
                                                             '[ -d /etc/cron.d ] && find /etc/cron.d ! -name \".*\" -type f -exec cat \{\} \;'])
     if exit_code != 0:
-        return '', None
-    return output.decode('utf8').replace('\t', ' '), False
-
-
-def get_container_crontab(channel: grpc.Channel, container_id: str) -> Tuple[str, Union[str, bool]]:
-    os_id = get_os_id(channel, container_id)
-    if os_id == 'alpine':
-        return get_alpine_crontab(channel, container_id)
-    elif os_id == 'debian' or os_id == 'ubuntu':
-        return get_debian_crontab(channel, container_id)
-    return '', None
+        return None
+    tab = output.decode('utf8').replace('\t', ' ')
+    return CronTab(tab=tab, user=False)
 
 
 def parse_args(command: str):
     return shlex.split(command)
 
 
-def get_container_name(spec):
+def get_container_name(spec, default: str):
+    if 'process' in spec:
+        process = spec['process']
+        if 'env' in process:
+            hostname_var: str = next(
+                filter(lambda v: v.startswith('HOSTNAME='), process['env']), None)
+            if hostname_var:
+                _, hostname = hostname_var.split('=', 2)
+                if hostname:
+                    return hostname
     if 'annotations' in spec:
         annotations = spec['annotations']
         if 'io.kubernetes.cri.container-name' in annotations:
@@ -201,22 +207,13 @@ def get_container_name(spec):
         return spec['hostname']
     if 'id' in spec:
         return spec['id']
+    return default
 
 
-def load_container_schedules(scheduler: BaseScheduler, container_id, channel):
-    tab, user = get_container_crontab(channel, container_id)
-    if tab == '':
-        return
-
+def crontab_to_schedule(channel: grpc.Channel, container_id: str, crontab: CronTab, scheduler: BaseScheduler):
     logger = logging.getLogger('cron')
-    container_spec = get_container_spec(channel, container_id)
-    container_name = get_container_name(
-        container_spec) if not None else container_id
-    logger.info(
-        'load schedules from [{container_name}]...'.format(container_name=container_name))
-
-    cron_jobs = CronTab(tab=tab, user=user)
-    for job in cron_jobs:
+    added = 0
+    for job in crontab:
         if not job.is_enabled():
             continue
         slices = str(job.slices)
@@ -228,13 +225,40 @@ def load_container_schedules(scheduler: BaseScheduler, container_id, channel):
                                 parse_args(job.command)],
                           name=job.command)
         logger.debug(
-            'found [{container_name}]: {job}.'.format(job=job.command, container_name=container_name))
+            'found {job}.'.format(job=job.command))
+        added += 1
+    return added
+
+
+def load_container_schedules(channel: grpc.Channel, container_id: str, scheduler: BaseScheduler):
+    logger = logging.getLogger('cron')
+    container_spec = get_container_spec(channel, container_id)
+    container_name = get_container_name(container_spec, container_id)
 
     logger.info(
-        'got {job_count} schedules now.'.format(job_count=len(scheduler.get_jobs())))
+        'load schedules from [{container_name}]...'.format(container_name=container_name))
+
+    added = 0
+    user = get_current_user(channel, container_id)
+    user_crontab = get_user_crontab(channel, container_id, user)
+    if user_crontab:
+        added += crontab_to_schedule(channel,
+                                     container_id, user_crontab, scheduler)
+
+    system_crontab = get_system_crontab(channel, container_id)
+    if system_crontab:
+        added += crontab_to_schedule(channel,
+                                     container_id, system_crontab, scheduler)
+
+    if added == 0:
+        logger.info('schedule not found in [{container_name}]'.format(
+            container_name=container_name))
+    else:
+        logger.info(
+            'got {job_count} schedules now.'.format(job_count=len(scheduler.get_jobs())))
 
 
-def unload_container_schedules(scheduler: BaseScheduler, container_id):
+def unload_container_schedules(channel: grpc.Channel, container_id: str, scheduler: BaseScheduler):
     jobs = scheduler.get_jobs()
     job_count = len(jobs)
     for job in jobs:
@@ -247,9 +271,11 @@ def unload_container_schedules(scheduler: BaseScheduler, container_id):
 
 
 def main():
+    global METADATA, FIFO_DIR
+
     try:
         opts, _ = getopt.getopt(sys.argv[1:], 's:n:', [
-            'cri-socket=', 'namespace='])
+            'cri-socket=', 'namespace=', 'fifo-dir='])
     except getopt.GetoptError as e:
         print('Usage: --cri-socket|-s <SOCKET> --namespace|-n <NAMESPACE>')
         exit(1)
@@ -262,12 +288,14 @@ def main():
             cri_socket = 'unix://' + v if v.startswith('/') else v
         elif k == '--namespace' or k == '-n':
             namespace = v
+        elif k == '--fifo-dir':
+            FIFO_DIR = v
 
-    global METADATA
     METADATA = (('containerd-namespace', namespace),)
-
     TIMEZONE = os.getenv('TIMEZONE', 'Asia/Shanghai')
 
+    os.makedirs(FIFO_DIR, exist_ok=True)
+
     logging.basicConfig(stream=sys.stdout)
     logging.getLogger('apscheduler').setLevel(logging.ERROR)
     logging.getLogger('cron').setLevel(logging.INFO)
@@ -277,19 +305,19 @@ def main():
     scheduler.start()
 
     with grpc.insecure_channel(cri_socket) as channel:
-        containers_stub = containers_pb2_grpc.ContainersStub(channel)
-        containers = containers_stub.List(
-            containers_pb2.ListContainersRequest(), metadata=METADATA).containers
-        for container in containers:
-            load_container_schedules(scheduler, container.id, channel)
+        tasks_stub = tasks_pb2_grpc.TasksStub(channel)
+        tasks = tasks_stub.List(tasks_pb2.ListTasksRequest(
+            filter='{.status=RUNNING}'), metadata=METADATA).tasks
+        for task in tasks:
+            load_container_schedules(channel, task.id, scheduler)
 
         events_stub = events_pb2_grpc.EventsStub(channel)
         for ev in events_stub.Subscribe(events_pb2.SubscribeRequest()):
             v = unwrap(ev)
             if ev.event.type_url == 'containerd.events.TaskCreate':
-                load_container_schedules(scheduler, v.container_id, channel)
+                load_container_schedules(channel, v.container_id, scheduler)
             elif ev.event.type_url == 'containerd.events.TaskDelete':
-                unload_container_schedules(scheduler, v.container_id)
+                unload_container_schedules(channel, v.container_id, scheduler)
 
 
 if __name__ == "__main__":
